
import os, re
import pandas as pd
import subprocess
from subprocess import PIPE
from tqdm import tqdm

"""
# Repository for investigating SATD included in quantum computers
## Target of investigation
python files contained in ./dependents/packages

## Details
Detect SATD based on patterns of 64 keywords that are likely to be included in SATD.
Outputs all SATD and unique SATD to a CSV file separately.
"""

# 64 keywords that are likely to be included in SATD
SATD_KEY_WORDS = """hack
retarded
at a loss
stupid
remove this code
ugly
take care of
something's gone wrong
nuke
is problematic
may cause problems
hacky
unknown why we ever experience this
treat this as a soft error
silly
workaround for bug
kludge
fixme
this isn't quite right
trial and error
give up
this is wrong
hang our heads in shame
temporary solution
causes issue
something bad is going on
cause for issue
this doesn't look right
is this next line safe
this indicates a more fundamental problem
temporary crutch
this can be a mess
this isn't very solid
this is temporary and will go away
is this line really safe
there is a problem
some fatal error
something serious is wrong
don't use this
get rid of this
doubt that this would work
this is bs
give up and go away
risk of this blowing up
just abandon it
prolly a bug
probably a bug
hope everything will work
toss it
barf
something bad happened
fix this crap
yuck
certainly buggy
remove me before production
you can be unhappy now
this is uncool
bail out
it doesn't work yet
crap
inconsistency
abandon all hope
kaboom
todo"""

# Comment detection
# Detect single-line comments and multi-line comments.
def detect_comment(txt):
    comments = re.findall(r"(([ \t\f\r]*#[^\n]*\n)+)", txt)
    comments = list(map(lambda tup: tup[0][:-1], comments)) # [:-1] is used to remove the final newline.
    comments.extend(re.findall(r'(\"{3}[^(\"{3})]*\"{3})', txt))
    comments.extend(re.findall(r"(\'{3}[^(\'{3})]*\'{3})", txt))

    # Remove the indentation before the comment.
    return ['\n'.join([cm.strip() for cm in comment.splitlines()]) for comment in comments]

# Determine whether it is SATD
# Determine whether the pattern of 64 keywords is included.
def is_satd(comment):
    return any(word in comment for word in SATD_KEY_WORDS.splitlines())


# Get SATD from the path to the file
# - Return SATD comments and all comments separately.
def get_satds_from_path(path):
    with open(path, 'r+') as f:
        content = f.read().lower()
        comments = detect_comment(content) # Comment detection
        return list(filter(lambda comment:is_satd(comment), comments)), comments


# Create a link to github
# - In the main branch, the file is updated as needed, so create a link to the file at a specific commit.
def get_link_to_file(repo):
    workingdir = f"./dependents/packages/{repo}"

    command = "git remote -v"
    proc = subprocess.run(command.split(), cwd=workingdir, stdout=PIPE, stderr=PIPE, text=True)
    """example result
    origin git@github.com:posl/quantum-computer-code-clone.git (fetch)
    origin git@github.com:posl/quantum-computer-code-clone.git (push)
    """
    owner = proc.stdout.split()[1].split(":")[1].split("/")[0]

    command = "git log -n 1 --format=%H"
    proc = subprocess.run(command.split(), cwd=workingdir, stdout=PIPE, stderr=PIPE, text=True)
    """example result
    ff834cea29ddc53edf232bff2acfb782acb22647
    """
    commit_id = proc.stdout.split()[0]

    return f"https://github.com/{owner}/{repo}/blob/{commit_id}/"


# The script that is executed first.
def main():
    """
    Format of CSV to be output
    - Repository name, path from repository to file, SATD, link to GitHub (link to file)
    """
    result = { "id": [], "repository": [], "path": [], "satd": [], "link to file": [] }
    result_unique = { "id": [], "repository": [], "path": [], "satd": [], "link to file": [] }
    id = 0

    # Holds the number of SATDs and comments.
    cnt = { "satds": 0, "comments": 0 }

    for dir in tqdm(os.listdir("./dependents/packages"), desc="Repository"):
        if dir == ".DS_Store": continue

        # Create a link to GitHub for each directory.
        link = get_link_to_file(dir)


        for cur_dir, _, files in os.walk(f"./dependents/packages/{dir}"):
            for file in files:
                if not file.endswith(".py"): continue

                pwd = cur_dir + "/" + file
                path = "/".join(pwd.split("/")[4:])
                satds, comments = get_satds_from_path(pwd)

                cnt["satds"] += len(satds)
                cnt["comments"] += len(comments)


                for satd in satds:
                    id += 1
                    # All SATDs
                    result["id"].append(f"ID{id}")
                    result["repository"].append(dir)
                    result["path"].append(path)
                    result["satd"].append(satd)
                    result["link to file"].append(link+path)

                    # Unique SATD
                    if not satd in result_unique["satd"]:
                        result_unique["id"].append(f"ID{id}")
                        result_unique["repository"].append(dir)
                        result_unique["path"].append(path)
                        result_unique["satd"].append(satd)
                        result_unique["link to file"].append(link+path)

    os.makedirs("./satd", exist_ok=True)
    # Output the results to a CSV file.
    df = pd.DataFrame.from_dict(result)
    df.set_index('id', inplace=True)
    df.to_csv("./satd/satd.csv")
    
    df_unique = pd.DataFrame.from_dict(result_unique)
    df_unique.set_index('id', inplace=True)
    df_unique.to_csv("./satd/satd_unique.csv")

    # Print the number of detected comments and SATD.
    print("=========================")
    print(f"comments: {cnt['comments']}")
    print(f"satds: {cnt['satds']}")
    print("satd.csv: ", len(df))
    print("satd_unique.csv: ", len(df_unique))


if __name__ == '__main__':
    main()